# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

from typing import Iterable, List, Optional, Tuple, TypeVar

import torch

from torcheval.metrics.functional.classification.precision_recall_curve import (
    _binary_precision_recall_curve_compute,
    _binary_precision_recall_curve_update,
    _multiclass_precision_recall_curve_compute,
    _multiclass_precision_recall_curve_update,
)
from torcheval.metrics.metric import Metric

"""
This file contains BinaryPrecisionRecallCurve and MulticlassPrecisionRecallCurve classes.
"""

TBinaryPrecisionRecallCurve = TypeVar("TBinaryPrecisionRecallCurve")
TMulticlassPrecisionRecallCurve = TypeVar("TMulticlassPrecisionRecallCurve")


class BinaryPrecisionRecallCurve(Metric[torch.Tensor]):
    """
    Compute precision recall curve, which is precision-recall pair with corresponding thresholds,
        for binary classification tasks.
    Its functional version is ``torcheval.metrics.functional.binary_precision_recall_curve``.

    Example:
        >>> import torch
        >>> from torcheval.metrics import BinaryPrecisionRecallCurve
        >>> metric = BinaryPrecisionRecallCurve()
        >>> input = torch.tensor([0.1, 0.5, 0.7, 0.8])
        >>> target = torch.tensor([0, 0, 1, 1])
        >>> metric.update(input, target)
        >>> metric.compute()
        (tensor([1., 1., 1.]),
        tensor([1.0000, 0.5000, 0.0000]),
        tensor([0.7000, 0.8000]))
    """

    def __init__(
        self: TBinaryPrecisionRecallCurve,
    ) -> None:
        super().__init__()
        self._add_state("inputs", [])
        self._add_state("targets", [])

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(
        self: TBinaryPrecisionRecallCurve,
        input: torch.Tensor,
        target: torch.Tensor,
    ) -> TBinaryPrecisionRecallCurve:
        """
        Update states with the ground truth labels and predictions.

        Args:
            input: Tensor of label predictions
                It should be probabilities or logits with shape of (n_sample, ).
            target: Tensor of ground truth labels with shape of (n_samples, ).
        """
        _binary_precision_recall_curve_update(
            input,
            target,
        )
        self.inputs.append(input)
        self.targets.append(target)
        return self

    @torch.inference_mode()
    # pyre-ignore[15]: compute() return tuple of precision, recall and thresholds
    def compute(
        self: TBinaryPrecisionRecallCurve,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Return:
            a tuple of (precision: torch.Tensor, recall: torch.Tensor, thresholds: torch.Tensor)
                precision: Tensor of precision result. Its shape is (n_thresholds + 1, )
                recall: Tensor of recall result. Its shape is (n_thresholds + 1, )
                thresholds: Tensor of threshold. Its shape is (n_thresholds, )
        """
        return _binary_precision_recall_curve_compute(
            torch.cat(self.inputs), torch.cat(self.targets)
        )

    @torch.inference_mode()
    def merge_state(
        self: TBinaryPrecisionRecallCurve,
        metrics: Iterable[TBinaryPrecisionRecallCurve],
    ) -> TBinaryPrecisionRecallCurve:
        for metric in metrics:
            if metric.inputs:
                metric_inputs = torch.cat(metric.inputs).to(self.device)
                metric_targets = torch.cat(metric.targets).to(self.device)
                self.inputs.append(metric_inputs)
                self.targets.append(metric_targets)
        return self

    @torch.inference_mode()
    def _prepare_for_merge_state(self: TBinaryPrecisionRecallCurve) -> None:
        if self.inputs and self.targets:
            self.inputs = [torch.cat(self.inputs)]
            self.targets = [torch.cat(self.targets)]


class MulticlassPrecisionRecallCurve(Metric[torch.Tensor]):
    """
    Compute precision recall curve, which is precision-recall pair with corresponding thresholds,
        for multi-class classification tasks.
    Its functional version is ``torcheval.metrics.functional.multiclass_precision_recall_curve``.

    Args:
        num_classes (Optional):
            Number of classes. Set to the second dimension of the input if num_classes is None.

    Example:
        >>> import torch
        >>> from torcheval.metrics import MulticlassPrecisionRecallCurve
        >>> metric = MulticlassPrecisionRecallCurve(num_classes=4)
        >>> input = torch.tensor([[0.1, 0.1, 0.1, 0.1], [0.5, 0.5, 0.5, 0.5], [0.7, 0.7, 0.7, 0.7], [0.8, 0.8, 0.8, 0.8]])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> metric.update(input, target)
        >>> metric.compute()
        ([tensor([0.2500, 0.0000, 0.0000, 0.0000, 1.0000]),
        tensor([0.3333, 0.0000, 0.0000, 1.0000]),
        tensor([0.5000, 0.0000, 1.0000]),
        tensor([1., 1.])],
        [tensor([1., 0., 0., 0., 0.]),
        tensor([1., 0., 0., 0.]),
        tensor([1., 0., 0.]),
        tensor([1., 0.])],
        [tensor([0.1000, 0.5000, 0.7000, 0.8000]),
        tensor([0.5000, 0.7000, 0.8000]),
        tensor([0.7000, 0.8000]),
        tensor([0.8000])])
    """

    def __init__(
        self: TMulticlassPrecisionRecallCurve,
        num_classes: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.num_classes = num_classes
        self._add_state("inputs", [])
        self._add_state("targets", [])

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(
        self: TMulticlassPrecisionRecallCurve,
        input: torch.Tensor,
        target: torch.Tensor,
    ) -> TMulticlassPrecisionRecallCurve:
        """
        Update states with the ground truth labels and predictions.

        Args:
            input: Tensor of label predictions
                It should be probabilities or logits with shape of (n_sample, n_class).
            target: Tensor of ground truth labels with shape of (n_samples, ).
        """
        _multiclass_precision_recall_curve_update(
            input,
            target,
            self.num_classes,
        )
        self.inputs.append(input)
        self.targets.append(target)
        return self

    @torch.inference_mode()
    # pyre-ignore[15]: compute() return tuple of precision, recall and thresholds
    def compute(
        self: TMulticlassPrecisionRecallCurve,
    ) -> Tuple[List[torch.Tensor], List[torch.Tensor], List[torch.Tensor]]:
        """
        Return:
            a tuple of (precision: List[torch.Tensor], recall: List[torch.Tensor], thresholds: List[torch.Tensor])
                precision: List of precision result. Each index indicates the result of a class.
                recall: List of recall result. Each index indicates the result of a class.
                thresholds: List of threshold. Each index indicates the result of a class.
        """
        return _multiclass_precision_recall_curve_compute(
            torch.cat(self.inputs), torch.cat(self.targets), self.num_classes
        )

    @torch.inference_mode()
    def merge_state(
        self: TMulticlassPrecisionRecallCurve,
        metrics: Iterable[TMulticlassPrecisionRecallCurve],
    ) -> TMulticlassPrecisionRecallCurve:
        for metric in metrics:
            if metric.inputs:
                metric_inputs = torch.cat(metric.inputs).to(self.device)
                metric_targets = torch.cat(metric.targets).to(self.device)
                self.inputs.append(metric_inputs)
                self.targets.append(metric_targets)
        return self

    @torch.inference_mode()
    def _prepare_for_merge_state(self: TMulticlassPrecisionRecallCurve) -> None:
        if self.inputs and self.targets:
            self.inputs = [torch.cat(self.inputs)]
            self.targets = [torch.cat(self.targets)]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import torch
from torch.nn import functional as F


@torch.inference_mode()
def binary_auroc(
    input: torch.Tensor,
    target: torch.Tensor,
) -> torch.Tensor:
    """
    Compute AUROC, which is the area under the ROC Curve, for binary classification.
    Its class version is ``torcheval.metrics.BinaryAUROC``.

    Args:
        input: Tensor of label predictions
            It should be predicted label, probabilities or logits with shape of (n_sample, ),
        target: Tensor of ground truth labels with shape of (n_samples, ).

    Example:
        >>> import torch
        >>> from torcheval.metrics.functional import binary_auroc
        >>> input = torch.tensor([0.1, 0.5, 0.7, 0.8])
        >>> target = torch.tensor([1, 0, 1, 1])
        >>> binary_auroc(input, target)
        tensor(0.6667)

        >>> input = torch.tensor([1, 1, 1, 0])
        >>> target = torch.tensor([1, 0, 1, 0])
        >>> binary_auroc(input, target)
        tensor(0.7500)
    """
    _auroc_update(input, target)
    return _auroc_compute(input, target)


def _auroc_update(
    input: torch.Tensor,
    target: torch.Tensor,
) -> None:
    _auroc_update_input_check(input, target)


@torch.jit.script
def _auroc_compute(
    input: torch.Tensor,
    target: torch.Tensor,
) -> torch.Tensor:
    threshold, indices = input.sort(descending=True)
    mask = F.pad(threshold.diff(dim=0) != 0, [0, 1], value=1.0)
    cum_tp = F.pad(target[indices].cumsum(0)[mask], pad=[1, 0], value=0.0)
    cum_fp = F.pad((1 - target[indices]).cumsum(0)[mask], pad=[1, 0], value=0.0)

    # Set AUROC to 0.5 when the target contains all ones or all zeros.
    factor = cum_tp[-1] * cum_fp[-1]
    auroc = torch.where(
        factor == 0,
        0.5,
        torch.trapz(cum_tp, cum_fp).double() / factor,
    )
    return auroc


def _auroc_update_input_check(
    input: torch.Tensor,
    target: torch.Tensor,
) -> None:
    if input.ndim != 1:
        raise ValueError(
            "input should be a one-dimensional tensor, " f"got shape {input.shape}."
        )
    if target.ndim != 1:
        raise ValueError(
            "target should be a one-dimensional tensor, " f"got shape {target.shape}."
        )
    if input.shape != target.shape:
        raise ValueError(
            "The `input` and `target` should have the same shape, "
            f"got shapes {input.shape} and {target.shape}."
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

from typing import Iterable, TypeVar

import torch

from torcheval.metrics.functional.regression.r2_score import (
    _r2_score_compute,
    _r2_score_param_check,
    _r2_score_update,
)
from torcheval.metrics.metric import Metric

TR2Score = TypeVar("TR2Score")


class R2Score(Metric[torch.Tensor]):
    """
    Compute R-squared score, which is the proportion of variance in the dependent variable that can be explained by the independent variable.
    Its functional version is ``torcheval.metrics.functional.r2_score``.

    Args:
        multioutput (Optional):
            - ``'uniform_average'``[default]:
                Return scores of all outputs are averaged with uniform weight.
            - ``'raw_values'``:
                Return a full set of scores.
            - ``variance_weighted``:
                Return scores of all outputs are averaged with weighted by the variances of each individual output.
        num_regressors (Optional):
            Number of independent variables used, applied to adjusted R-squared score. Defaults to zero (standard R-squared score).
    Raises:
        ValueError:
            - If value of multioutput does not exist in (``raw_values``, ``uniform_average``, ``variance_weighted``).
            - If value of num_regressors is not an ``integer`` in the range of [0, n_samples - 1].

    Example:
        >>> import torch
        >>> from torcheval.metrics import R2Score
        >>> metric = R2Score()
        >>> input = torch.tensor([0, 2, 1, 3])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.6)

        >>> metric = R2Score()
        >>> input = torch.tensor([[0, 2], [1, 6]])
        >>> target = torch.tensor([[0, 1], [2, 5]])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.6250)

        >>> metric = R2Score(multioutput="raw_values")
        >>> input = torch.tensor([[0, 2], [1, 6]])
        >>> target = torch.tensor([[0, 1], [2, 5]])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor([0.5000, 0.7500])

        >>> metric = R2Score(multioutput="variance_weighted")
        >>> input = torch.tensor([[0, 2], [1, 6]])
        >>> target = torch.tensor([[0, 1], [2, 5]])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.7000)

        >>> metric = R2Score(multioutput="raw_values", num_regressors=2)
        >>> input = torch.tensor([1.2, 2.5, 3.6, 4.5, 6])
        >>> target = torch.tensor([1, 2, 3, 4, 5])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.6200)
    """

    def __init__(
        self: TR2Score,
        multioutput: str = "uniform_average",
        num_regressors: int = 0,
    ) -> None:
        super().__init__()
        _r2_score_param_check(multioutput, num_regressors)
        self.multioutput = multioutput
        self.num_regressors = num_regressors
        self._add_state("sum_squared_obs", torch.tensor(0.0))
        self._add_state("sum_obs", torch.tensor(0.0))
        self._add_state("sum_squared_residual", torch.tensor(0.0))
        self._add_state("num_obs", torch.tensor(0.0))

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(
        self: TR2Score,
        input: torch.Tensor,
        target: torch.Tensor,
    ) -> TR2Score:
        """
        Update states with the ground truth values and predictions.

        Args:
            input: Tensor of predicted values with shape of (n_sample, n_output).
            target: Tensor of ground truth values with shape of (n_sample, n_output).
        """
        (
            sum_squared_obs,
            sum_obs,
            sum_squared_residual,
            num_obs,
        ) = _r2_score_update(input, target)
        if self.sum_squared_obs.ndim == 0 and sum_squared_obs.ndim == 1:
            self.sum_squared_obs = sum_squared_obs
            self.sum_obs = sum_obs
            self.sum_squared_residual = sum_squared_residual
        else:
            self.sum_squared_obs += sum_squared_obs
            self.sum_obs += sum_obs
            self.sum_squared_residual += sum_squared_residual
        self.num_obs += num_obs
        return self

    @torch.inference_mode()
    def compute(self: TR2Score) -> torch.Tensor:
        """
        Return the R-squared score.

        NaN is returned if no calls to ``update()`` are made before ``compute()`` is called.
        """
        return _r2_score_compute(
            self.sum_squared_obs,
            self.sum_obs,
            self.sum_squared_residual,
            self.num_obs,
            self.multioutput,
            self.num_regressors,
        )

    @torch.inference_mode()
    def merge_state(self: TR2Score, metrics: Iterable[TR2Score]) -> TR2Score:
        for metric in metrics:
            if self.sum_squared_obs.ndim == 0 and metric.sum_squared_obs.ndim == 1:
                self.sum_squared_obs = metric.sum_squared_obs.to(self.device)
                self.sum_obs = metric.sum_obs.to(self.device)
                self.sum_squared_residual = metric.sum_squared_residual.to(self.device)
            else:
                self.sum_squared_obs += metric.sum_squared_obs.to(self.device)
                self.sum_obs += metric.sum_obs.to(self.device)
                self.sum_squared_residual += metric.sum_squared_residual.to(self.device)
            self.num_obs += metric.num_obs.to(self.device)
        return self



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from torcheval.metrics.regression.mean_squared_error import MeanSquaredError
from torcheval.metrics.regression.r2_score import R2Score

__all__ = ["MeanSquaredError", "R2Score"]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

from typing import Iterable, Optional, TypeVar

import torch

from torcheval.metrics.functional.regression.mean_squared_error import (
    _mean_squared_error_compute,
    _mean_squared_error_param_check,
    _mean_squared_error_update,
)
from torcheval.metrics.metric import Metric

TMeanSquaredError = TypeVar("TMeanSquaredError")


class MeanSquaredError(Metric[torch.Tensor]):
    """
    Compute Mean Squared Error, which is the mean of squared error of `input` and `target`.
    Its functional version is ``torcheval.metrics.functional.mean_squared_error``.

    Args:
        multioutput (Optional):
            - ``'uniform_average'``[default]:
                Return scores of all outputs are averaged with uniform weight.
            - ``'raw_values'``:
                Return a full set of scores.
    Raises:
        ValueError:
            - If value of multioutput does not exist in (``raw_values``, ``uniform_average``).
            - If the dimension of `input` or `target` is not 1D or 2D.
            - If the `input` and `target` do not have the same size.
            - If the first dimension of `input`, `target` and `sample_weight` are not the same.

    Example:
        >>> import torch
        >>> from torcheval.metrics import MeanSquaredError
        >>> metric = MeanSquaredError()
        >>> input = torch.tensor([0.9, 0.5, 0.3, 0.5])
        >>> target = torch.tensor([0.5, 0.8, 0.2, 0.8])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.0875)

        >>> metric = MeanSquaredError()
        >>> input = torch.tensor([[0.9, 0.5], [0.3, 0.5]])
        >>> target = torch.tensor([[0.5, 0.8], [0.2, 0.8]])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.0875)

        >>> metric = MeanSquaredError(multioutput="raw_values")
        >>> input = torch.tensor([[0.9, 0.5], [0.3, 0.5]])
        >>> target = torch.tensor([[0.5, 0.8], [0.2, 0.8]])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor([0.0850, 0.0900])

        >>> input = torch.tensor([[0.9, 0.5], [0.3, 0.5]])
        >>> target = torch.tensor([[0.5, 0.8], [0.2, 0.8]])
        >>> metric.update(input, target, sample_weight=torch.tensor([0.2, 0.8]))
        >>> metric.compute()
        tensor(0.0650)
    """

    def __init__(
        self: TMeanSquaredError,
        multioutput: str = "uniform_average",
    ) -> None:
        super().__init__()
        _mean_squared_error_param_check(multioutput)
        self.multioutput = multioutput
        self._add_state("sum_squared_error", torch.tensor(0.0))
        self._add_state("sum_weight", torch.tensor(0.0))

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(
        self: TMeanSquaredError,
        input: torch.Tensor,
        target: torch.Tensor,
        sample_weight: Optional[torch.Tensor] = None,
    ) -> TMeanSquaredError:
        """
        Update states with the ground truth values and predictions.

        Args:
            input: Tensor of predicted values with shape of (n_sample, n_output).
            target: Tensor of ground truth values with shape of (n_sample, n_output).
            sample_weight (Optional):
                Tensor of sample weights with shape of (n_sample, ). Defaults to None.
        """
        (
            sum_squared_error,
            sum_weight,
        ) = _mean_squared_error_update(input, target, sample_weight)
        if self.sum_squared_error.ndim == 0 and sum_squared_error.ndim == 1:
            self.sum_squared_error = sum_squared_error
        else:
            self.sum_squared_error += sum_squared_error
        self.sum_weight += sum_weight
        return self

    @torch.inference_mode()
    def compute(self: TMeanSquaredError) -> torch.Tensor:
        """
        Return the Mean Squared Error.

        NaN is returned if no calls to ``update()`` are made before ``compute()`` is called.
        """
        return _mean_squared_error_compute(
            self.sum_squared_error,
            self.multioutput,
            self.sum_weight,
        )

    @torch.inference_mode()
    def merge_state(
        self: TMeanSquaredError, metrics: Iterable[TMeanSquaredError]
    ) -> TMeanSquaredError:
        for metric in metrics:
            if self.sum_squared_error.ndim == 0 and metric.sum_squared_error.ndim == 1:
                self.sum_squared_error = metric.sum_squared_error.to(self.device)
            else:
                self.sum_squared_error += metric.sum_squared_error.to(self.device)
            self.sum_weight += metric.sum_weight.to(self.device)
        return self



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import os
import sys

from datetime import date

from typing import List

from setuptools import find_packages, setup
from torcheval import __version__


def current_path(file_name: str) -> str:
    return os.path.abspath(os.path.join(__file__, os.path.pardir, file_name))


def read_requirements(file_name: str) -> List[str]:
    with open(current_path(file_name), encoding="utf8") as f:
        return f.read().strip().split()


def get_nightly_version() -> str:
    return date.today().strftime("%Y.%m.%d")


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="torcheval setup")
    parser.add_argument(
        "--nightly",
        dest="nightly",
        action="store_true",
        help="enable settings for nightly package build",
    )
    parser.set_defaults(nightly=False)
    return parser.parse_known_args()


if __name__ == "__main__":
    with open(current_path("README.md"), encoding="utf8") as f:
        readme = f.read()

    custom_args, setup_args = parse_args()
    package_name = "torcheval" if not custom_args.nightly else "torcheval-nightly"
    version = __version__ if not custom_args.nightly else get_nightly_version()
    print(f"using package_name={package_name}, version={version}")

    sys.argv = [sys.argv[0]] + setup_args

    setup(
        name=package_name,
        version=version,
        author="torcheval team",
        author_email="yicongd@fb.com",
        description="A library for providing a simple interface to create new metrics and an easy-to-use toolkit for metric computations and checkpointing.",
        long_description=readme,
        long_description_content_type="text/markdown",
        url="https://github.com/facebookresearch/torcheval",
        license="BSD-3",
        keywords=["pytorch", "evaluation", "metrics"],
        python_requires=">=3.7",
        install_requires=read_requirements("requirements.txt"),
        packages=find_packages(),
        zip_safe=True,
        classifiers=[
            "Development Status :: 2 - Pre-Alpha",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: BSD License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
        ],
        extras_require={"dev": read_requirements("dev-requirements.txt")},
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

from typing import Iterable, TypeVar

import torch

from torcheval.metrics.metric import Metric


TMax = TypeVar("TMax")


class Max(Metric[torch.Tensor]):
    """
    Calculate the maximum value of all elements in all the input tensors.
    Its functional version is ``torch.max(input)``.

    Example:
        >>> import torch
        >>> from torcheval.metrics import Max
        >>> metric = Max()
        >>> metric.update(torch.tensor([[1, 2], [3, 4]]))
        >>> metric.compute()
        tensor(4.)

        >>> metric.update(torch.tensor(-1)).compute()
        tensor(4.)

        >>> metric.reset()
        >>> metric.update(torch.tensor(-1)).compute()
        tensor(-1.)
    """

    def __init__(self: TMax) -> None:
        super().__init__()
        self._add_state("max", torch.tensor(float("-inf")))

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(self: TMax, input: torch.Tensor) -> TMax:
        self.max = torch.max(self.max, torch.max(input))
        return self

    @torch.inference_mode()
    def compute(self: TMax) -> torch.Tensor:
        return self.max

    @torch.inference_mode()
    def merge_state(self: TMax, metrics: Iterable[TMax]) -> TMax:
        for metric in metrics:
            self.max = torch.max(self.max, metric.max.to(self.device))
        return self



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[6]: expect int got Optional[int] for num_classes
# pyre-ignore-all-errors[58]: - is not supported for operand types Optional[int] and int

import logging
from typing import Optional, Tuple

import torch


@torch.inference_mode()
def multiclass_precision(
    input: torch.Tensor,
    target: torch.Tensor,
    num_classes: Optional[int] = None,
    average: Optional[str] = "micro",
) -> torch.Tensor:
    """
    Compute precision score, which is the ratio of the true positives (TP) and the
    total number of points classified as positives (TP + FP).
    Its class version is ``torcheval.metrics.MultiClassPrecision``.

    Args:
        input: Tensor of label predictions
            It could be the predicted labels, with shape of (n_sample, ).
            It could also be probabilities or logits with shape of (n_sample, n_class).
            ``torch.argmax`` will be used to convert input into predicted labels.
        target: Tensor of ground truth labels with shape of (n_sample, ).
        num_classes:
            Number of classes.
        average:
            - ``'micro'``[default]:
                Calculate the metrics globally, by using the total true positives and false
                positives across all classes.
            - ``'macro'``:
                Calculate metrics for each class separately, and return their unweighted
                mean. Classes with 0 true instances and predicted instances are ignored.
            - ``'weighted'``:
                Calculate metrics for each class separately, and return their average weighted
                by the number of instances for each class in the ``target`` tensor. Classes with
                0 true instances and predicted instances are ignored.
            - ``None``:
                Calculate the metric for each class separately, and return
                the metric for every class.
                NaN is returned if a class has no sample in ``target``.

    Example:
        >>> import torch
        >>> from torcheval.metrics.functional import multiclass_precision
        >>> input = torch.tensor([0, 2, 1, 3])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> multiclass_precision(input, target)
        tensor(0.5)
        >>> multiclass_precision(input, target, average=None, num_classes=4)
        tensor([1., 0., 0., 1.])
        >>> multiclass_precision(input, target, average="macro", num_classes=4)
        tensor(0.5)
        >>> input = torch.tensor([[0.9, 0.1, 0, 0], [0.1, 0.2, 0.4, 0,3], [0, 1.0, 0, 0], [0, 0, 0.2, 0.8]])
        >>> multiclass_precision(input, target)
        tensor(0.5)
    """

    _precision_param_check(num_classes, average)
    num_tp, num_fp, num_class = _precision_update(input, target, num_classes, average)
    return _precision_compute(num_tp, num_fp, num_class, average)


def _precision_update(
    input: torch.Tensor,
    target: torch.Tensor,
    num_classes: Optional[int],
    average: Optional[str],
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    _precision_update_input_check(input, target, num_classes)

    if input.ndim == 2:
        input = torch.argmax(input, dim=1)

    if average == "micro":
        num_tp = (input == target).sum()
        num_fp = (input != target).sum()
        return num_tp, num_fp, torch.tensor(0.0)

    target_ = target.to(dtype=torch.float32)
    input_ = input.to(dtype=torch.float32)

    num_label = torch.histc(target_, bins=num_classes, min=0, max=num_classes - 1)
    num_tp = torch.histc(
        input_[input == target], bins=num_classes, min=0, max=num_classes - 1
    )
    num_fp = torch.histc(
        input_[input != target], bins=num_classes, min=0, max=num_classes - 1
    )

    return num_tp, num_fp, num_label


def _precision_compute(
    num_tp: torch.Tensor,
    num_fp: torch.Tensor,
    num_label: torch.Tensor,
    average: Optional[str],
) -> torch.Tensor:

    if average in ("macro", "weighted"):

        # Ignore the class that has no samples in both `input` and `target`
        mask = (num_label != 0) | ((num_tp + num_fp) != 0)
        num_tp, num_fp = num_tp[mask], num_fp[mask]

    precision = num_tp / (num_tp + num_fp)

    if average in (None, "None") and torch.isnan(precision).sum():
        # Warning in case there are classes with zero representation in both the
        # predictions and the ground truth.
        bad_class = torch.nonzero(torch.isnan(precision))
        logging.warning(
            f"{bad_class} classes have zero instances in both the "
            "predictions and the ground truth labels. Precision is still logged "
            "as zero."
        )

    # If precision is NaN, convert it to 0.
    precision = torch.nan_to_num(precision)

    if average == "micro":
        return precision
    elif average == "macro":
        return precision.mean()
    elif average == "weighted":
        return torch.inner(precision, (num_label[mask] / num_label.sum()))
    else:  # average is None
        return precision


def _precision_param_check(
    num_classes: Optional[int],
    average: Optional[str],
) -> None:
    average_options = ("micro", "macro", "weighted", "None", None)
    if average not in average_options:
        raise ValueError(
            f"`average` was not in the allowed value of {average_options}, got {average}."
        )
    if average != "micro" and (num_classes is None or num_classes <= 0):
        raise ValueError(
            f"num_classes should be a positive number when average={average}."
            f" Got num_classes={num_classes}."
        )


def _precision_update_input_check(
    input: torch.Tensor,
    target: torch.Tensor,
    num_classes: Optional[int],
) -> None:
    if input.size(0) != target.size(0):
        raise ValueError(
            "The `input` and `target` should have the same first dimension, "
            f"got shapes {input.shape} and {target.shape}."
        )

    if target.ndim != 1:
        raise ValueError(
            f"target should be a one-dimensional tensor, got shape {target.shape}."
        )

    if not input.ndim == 1 and not (
        input.ndim == 2 and (num_classes is None or input.shape[1] == num_classes)
    ):
        raise ValueError(
            "input should have shape of (num_sample,) or (num_sample, num_classes), "
            f"got {input.shape}."
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

from typing import Iterable, Optional, TypeVar

import torch
from torcheval.metrics.functional.classification.recall import (
    _recall_compute,
    _recall_param_check,
    _recall_update,
)
from torcheval.metrics.metric import Metric

TRecall = TypeVar("TRecall")


class MulticlassRecall(Metric[torch.Tensor]):
    """
    Compute the recall score, the ratio of the true positives and the sum of
    true positives and false negatives.
    Its functional version is ``torcheval.metrics.functional.multiclass_recall``.
    We cast NaNs to 0 when classes have zero instances in the ground-truth labels
    (when TP + FN = 0).

    Args:
        num_classes:
            Number of classes.
        average:
            - ``'micro'``[default]:
                Calculate the metrics globally.
            - ``'macro'``:
                Calculate metrics for each class separately, and return their unweighted mean.
                Classes with 0 true and predicted instances are ignored.
            - ``'weighted'``"
                Calculate metrics for each class separately, and return their weighted sum.
                Weights are defined as the proportion of occurrences of each class in "target".
                Classes with 0 true and predicted instances are ignored.
            - ``None``:
                Calculate the metric for each class separately, and return
                the metric for every class.

    Example:
        >>> import torch
        >>> from torcheval.metrics.classification import MulticlassRecall
        >>> metric = MulticlassRecall(num_classes=4)
        >>> input = torch.tensor([0, 2, 1, 3])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.5000)

        >>> metric = MulticlassRecall(average=None, num_classes=4)
        >>> input = torch.tensor([0, 2, 1, 3])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor([1., 0., 0., 1.])

        >>> metric = MulticlassRecall(average="macro", num_classes=2)
        >>> input = torch.tensor([0, 0, 1, 1, 1])
        >>> target = torch.tensor([0, 0, 0, 0, 1])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.5000)

        >>> metric = MulticlassRecall(num_classes=4)
        >>> input = torch.tensor([[0.9, 0.1, 0, 0], [0.1, 0.2, 0.4, 0.3], [0, 1.0, 0, 0], [0, 0, 0.2, 0.8]])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.5000)
    """

    def __init__(
        self: TRecall,
        num_classes: Optional[int] = None,
        average: Optional[str] = "micro",
    ) -> None:
        super().__init__()
        _recall_param_check(num_classes, average)
        self.num_classes = num_classes
        self.average = average

        if average == "micro":
            self._add_state("num_tp", torch.tensor(0.0))
            self._add_state("num_labels", torch.tensor(0.0))
            self._add_state("num_predictions", torch.tensor(0.0))
        else:
            assert isinstance(
                num_classes, int
            ), f"`num_classes` must be an integer, but got {num_classes}."
            self._add_state("num_tp", torch.zeros(num_classes))
            self._add_state("num_labels", torch.zeros(num_classes))
            self._add_state("num_predictions", torch.zeros(num_classes))

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(self: TRecall, input: torch.Tensor, target: torch.Tensor) -> TRecall:
        """
        Update states with the ground truth labels and predictions.

        Args:
            input: Tensor of label predictions.
                It could be the predicted labels, with shape of (n_sample, ).
                It could also be probabilities or logits with shape of (n_sample, n_class).
                ``torch.argmax`` will be used to convert input into predicted labels.
            target: Tensor of ground truth labels with shape of (n_sample, ).
        """
        num_tp, num_labels, num_predictions = _recall_update(
            input, target, self.num_classes, self.average
        )
        self.num_tp += num_tp
        self.num_labels += num_labels
        self.num_predictions += num_predictions
        return self

    @torch.inference_mode()
    def compute(self: TRecall) -> torch.Tensor:
        """
        Return the recall score.

        NaN is returned if no calls to ``update()`` are made before ``compute()`` is called.
        """
        return _recall_compute(
            self.num_tp, self.num_labels, self.num_predictions, self.average
        )

    @torch.inference_mode()
    def merge_state(self: TRecall, metrics: Iterable[TRecall]) -> TRecall:
        for metric in metrics:
            self.num_tp += metric.num_tp.to(self.device)
            self.num_labels += metric.num_labels.to(self.device)
            self.num_predictions += metric.num_predictions.to(self.device)
        return self



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

import logging
from typing import Iterable, TypeVar, Union

import torch

from torcheval.metrics.functional.aggregation.mean import _mean_update
from torcheval.metrics.metric import Metric

TMean = TypeVar("TMean")


class Mean(Metric[torch.Tensor]):
    """
    Calculate the weighted mean value of all elements in all the input tensors.
    When weight is not provided, it calculates the unweighted mean.
    Its functional version is ``torcheval.functional.mean()``.

    Example:
        >>> import torch
        >>> from torcheval.metrics import Mean
        >>> metric = Mean()
        >>> metric.update(1)
        >>> metric.update(torch.tensor([2, 3]))
        >>> metric.compute()
        tensor(2.)

        >>> metric.update(torch.tensor(-1)).compute()
        tensor(1.25)

        >>> metric.reset()
        >>> metric.update(torch.tensor(-1)).compute()
        tensor(-1.)

        >>> metric = Mean()
        >>> metric.update(torch.tensor([2, 3]), torch.tensor([0.2, 0.8])).compute()
        tensor(2.8)
        >>> metric.update(torch.tensor([4, 5]), 0.5).compute()
        tensor(3.65)
        >>> metric.update(torch.tensor([6]), 2).compute()
        tensor(4.825)
    """

    def __init__(self: TMean) -> None:
        super().__init__()
        self._add_state("weighted_sum", torch.tensor(0.0))
        self._add_state("weights", torch.tensor(0.0))

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(
        self: TMean, input: torch.Tensor, weight: Union[float, int, torch.Tensor] = 1.0
    ) -> TMean:
        """
        Compute weighted mean. When weight is not provided, it calculates the unweighted mean.

        weighted_mean = sum(weight * input) / sum(weight)

        Args:
            input: Tensor of input values.
            weight(optional): Float or Int or Tensor of input weights. It is default to 1.0. If weight is a Tensor, its size should match the input tensor size.
        Raises:
            ValueError: If value of weight is neither a ``float`` nor a ``int'' nor a ``torch.Tensor`` that matches the input tensor size.
        """

        weighted_sum, weights = _mean_update(input, weight)
        self.weighted_sum += weighted_sum
        self.weights += weights
        return self

    @torch.inference_mode()
    def compute(self: TMean) -> torch.Tensor:
        """
        If no calls to ``update()`` are made before ``compute()`` is called,
        the function throws a warning and returns 0.0.
        """
        if not self.weighted_sum:
            logging.warning("No calls to update() have been made - returning 0.0")
            return torch.tensor(0.0)
        return self.weighted_sum / self.weights

    @torch.inference_mode()
    def merge_state(self: TMean, metrics: Iterable[TMean]) -> TMean:
        for metric in metrics:
            self.weighted_sum += metric.weighted_sum.to(self.device)
            self.weights += metric.weights.to(self.device)
        return self



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from abc import ABC, abstractmethod
from collections import defaultdict
from copy import deepcopy
from typing import Any, Dict, Generic, Iterable, List, TypeVar, Union

import torch

# pyre-fixme[24]: Generic type `Metric` expects 1 type parameter.
TSelf = TypeVar("TSelf", bound="Metric")
TComputeReturn = TypeVar("TComputeReturn")
# pyre-ignore[33]: Flexible key data type for dictionary
TState = Union[torch.Tensor, List[torch.Tensor], Dict[Any, torch.Tensor]]


class Metric(Generic[TComputeReturn], ABC):
    """
    Base class for all metrics present in the Metrics API.

    Implement __init__(), update(), compute(), merge_state() functions
    to implement your own metric.
    """

    def __init__(self: TSelf) -> None:
        """
        Initialize a metric object and its internal states.

        Use ``self._add_state()`` to initialize state variables of your metric class.
        The state variables should be either ``torch.Tensor``, a list of
        ``torch.Tensor`` or a dictionary with ``torch.Tensor`` as values.
        """
        torch._C._log_api_usage_once(f"torcheval.metrics.{self.__class__.__name__}")

        # limit state variable type to tensor/[tensor] to avoid working with nested
        # data structures when move/detach/clone tensors. Can open more types up
        # upon user requests in the future.
        self._state_name_to_default: Dict[str, TState] = {}
        self._device: torch.device = torch.device("cpu")

    def _add_state(self: TSelf, name: str, default: TState) -> None:
        """
        Used in subclass ``__init__()`` to add a metric state variable.

        Args:
            name: The name of the state variable. The variable can be accessed
                with ``self.name``.
            default: Default value of the state. It can either be a ``torch.Tensor``,
                a list of ``torch.Tensor`` or a dictionary with ``torch.Tensor`` as values.
                The state will be reset to this value when ``self.reset()`` is called.
        Raises:
            TypeError: If ``default`` is not a ``torch.Tensor``, a list of
                ``torch.Tensor`` or a dictionary with ``torch.Tensor`` as values.
        """
        _check_state_variable_type(name, default)
        # deepcopy makes sure the input/initial value/default value of the state
        # variable are independent.
        setattr(self, name, deepcopy(default))
        self._state_name_to_default[name] = deepcopy(default)

    @abstractmethod
    @torch.inference_mode()
    def update(self: TSelf, *_: Any, **__: Any) -> TSelf:
        """
        Implement this method to update the state variables of your metric class.

        Decorate update() with @torch.inference_mode() which gives better
        performance by disabling view tracking.
        """

    @abstractmethod
    @torch.inference_mode()
    def compute(self: TSelf) -> TComputeReturn:
        """
        Implement this method to compute and return the final metric value
        from state variables.

        Decorate compute() with @torch.inference_mode() which gives better
        performance by disabling view tracking.
        """

    @abstractmethod
    @torch.inference_mode()
    def merge_state(self: TSelf, metrics: Iterable[TSelf]) -> TSelf:
        """
        Implement this method to update the current metric's state variables
        to be the merged states of the current metric and input metrics. The state
        variables of input metrics should stay unchanged.

        Decorate merge_state() with @torch.inference_mode() which gives better
        performance by disabling view tracking.

        ``self.merge_state`` might change the size/shape of state variables.
        Make sure ``self.update``and ``self.compute`` can still be called
        without exceptions when state variables are merged.

        This method can be used as a building block for syncing metric states
        in distributed training. For example, ``sync_and_compute`` in the metric
        toolkit will use this method to merge metric objects gathered from the
        process group.
        """

    @torch.inference_mode()
    def _prepare_for_merge_state(self: TSelf) -> None:
        """
        Called before syncing metrics in ``toolkit._sync_metric_object()``.

        It can be utilized to adjust metric states to accelerate syncing.
        For example, concatenated metric state from a list of tensors to
        one tensor. See ``torcheval.metrics.BinaryAUROC`` as an example.
        """
        pass

    def reset(self: TSelf) -> TSelf:
        """
        Reset the metric state variables to their default value.
        The tensors in the default values are also moved to the device of
        the last ``self.to(device)`` call.
        """
        for state_name, default in self._state_name_to_default.items():
            if isinstance(default, torch.Tensor):
                setattr(self, state_name, default.clone().to(self.device))
            elif isinstance(default, list):
                setattr(
                    self,
                    state_name,
                    [tensor.clone().to(self.device) for tensor in default],
                )
            elif isinstance(default, dict):
                setattr(
                    self,
                    state_name,
                    defaultdict(
                        lambda: torch.tensor(0.0, device=self.device),
                        {
                            key: tensor.clone().to(self.device)
                            for key, tensor in default.items()
                        },
                    ),
                )
        return self

    def state_dict(self: TSelf) -> Dict[str, TState]:
        """
        Save metric state variables in state_dict.

        Raises:
            TypeError: If ``default`` is not a ``torch.Tensor``, a list of
                ``torch.Tensor`` or a dictionary with ``torch.Tensor`` as values.
        """
        state_dict = {}
        for state_name in self._state_name_to_default:
            value = getattr(self, state_name)
            _check_state_variable_type(state_name, value)

            if isinstance(value, torch.Tensor):
                state_dict[state_name] = value.detach().clone()
            elif isinstance(value, list):
                state_dict[state_name] = [tensor.detach().clone() for tensor in value]
            elif isinstance(value, dict):
                state_dict[state_name] = {
                    key: tensor.detach().clone() for key, tensor in value.items()
                }
        return state_dict

    def load_state_dict(
        self: TSelf,
        state_dict: Dict[str, Any],
        strict: bool = True,
    ) -> None:
        """
        Loads metric state variables from state_dict.

        Args:
            state_dict: The name of the state variable.
            strict: Whether to strictly enforce that the keys in ``state_dict`` matches
                all names of the metric states.
        Raises:
            RuntimeError: If ``strict`` is ``True`` and keys in state_dict does not match
                all names of the metric states.
            TypeError: If ``default`` is not a ``torch.Tensor``, a list of
                ``torch.Tensor`` or a dictionary with ``torch.Tensor`` as values.
        """
        state_dict = deepcopy(state_dict)
        metric_state_names = set(self._state_name_to_default.keys())
        for state_name in metric_state_names:
            if state_name in state_dict:
                value = state_dict[state_name]
                _check_state_variable_type(state_name, value)
                setattr(self, state_name, value)

        if strict:
            state_dict_keys = set(state_dict.keys())
            unexpected_keys = state_dict_keys.difference(metric_state_names)
            missing_keys = metric_state_names.difference(state_dict_keys)
            if missing_keys or unexpected_keys:
                raise RuntimeError(
                    f"Error(s) in loading state_dict for {self.__class__.__name__}. "
                    f"Encountered missing keys: {missing_keys} and unexpected "
                    f"keys: {unexpected_keys}."
                )

    def to(
        self: TSelf, device: Union[str, torch.device], *args: Any, **kwargs: Any
    ) -> TSelf:
        """
        Move tensors in metric state variables to device.

        Args:
            device: The destination device.
        Raises:
            TypeError: If ``default`` is not a ``torch.Tensor``, a list of
                ``torch.Tensor`` or a dictionary with ``torch.Tensor`` as values.
        """
        device = torch.device(device) if isinstance(device, str) else device
        for state_name in self._state_name_to_default:
            value = getattr(self, state_name)
            _check_state_variable_type(state_name, value)
            if isinstance(value, torch.Tensor):
                setattr(self, state_name, value.to(device))
            elif isinstance(value, list):
                setattr(
                    self,
                    state_name,
                    [tensor.to(device, *args, **kwargs) for tensor in value],
                )
            elif isinstance(value, dict):
                setattr(
                    self,
                    state_name,
                    defaultdict(
                        lambda: torch.tensor(0.0, device=device),
                        {
                            key: tensor.to(device, *args, **kwargs)
                            for key, tensor in value.items()
                        },
                    ),
                )
        self._device = device
        return self

    @property
    def device(self: TSelf) -> torch.device:
        """
        The last input device of ``Metric.to()``.
        Default to ``torch.device("cpu")`` if ``Metric.to()`` is not called.
        """
        return self._device


# pyre-ignore[2]: Type checking for ``value`` which could be any type.
def _check_state_variable_type(name: str, value: Any) -> None:
    """
    Check the type of a state variable value.
    It should be either a ``torch.Tensor``, a list of ``torch.Tensor``,
    or a dictionary with ``torch.Tensor`` as values.
    """
    if (
        not isinstance(value, torch.Tensor)
        and not (
            isinstance(value, list) and all(isinstance(x, torch.Tensor) for x in value)
        )
        and not (
            isinstance(value, dict)
            and all(isinstance(x, torch.Tensor) for x in value.values())
        )
    ):
        raise TypeError(
            "The value of state variable must be a ``torch.Tensor``, a list of ``torch.Tensor``, "
            f"or a dictionary with ``torch.Tensor`` as values. "
            f"Get {name}={value} instead."
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

from typing import Tuple

import torch


@torch.inference_mode()
def r2_score(
    input: torch.Tensor,
    target: torch.Tensor,
    multioutput: str = "uniform_average",
    num_regressors: int = 0,
) -> torch.Tensor:
    """
    Compute R-squared score, which is the proportion of variance in the dependent variable that can be explained by the independent variable.
    Its class version is ``torcheval.metrics.R2Score``.

    Args:
        input:
            Tensor of predicted values with shape of (n_sample, n_output).
        target:
            Tensor of ground truth values with shape of (n_sample, n_output).
        multioutput (Optional):
            - ``'uniform_average'``[default]:
                Return scores of all outputs are averaged with uniform weight.
            - ``'raw_values'``:
                Return a full set of scores.
            - ``variance_weighted``:
                Return scores of all outputs are averaged with weighted by the variances of each individual output.
        num_regressors (Optional):
            Number of independent variables used, applied to adjusted R-squared score. Defaults to zero (standard R-squared score).
    Raises:
        ValueError:
            - If value of multioutput does not exist in (``raw_values``, ``uniform_average``, ``variance_weighted``).
            - If value of num_regressors is not an ``integer`` in the range of [0, n_samples - 1].

    Example:
        >>> import torch
        >>> from torcheval.metrics.functional import r2_score
        >>> input = torch.tensor([0, 2, 1, 3])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> r2_score(input, target)
        tensor(0.6)

        >>> input = torch.tensor([[0, 2], [1, 6]])
        >>> target = torch.tensor([[0, 1], [2, 5]])
        >>> r2_score(input, target)
        tensor(0.6250)

        >>> input = torch.tensor([[0, 2], [1, 6]])
        >>> target = torch.tensor([[0, 1], [2, 5]])
        >>> r2_score(input, target, multioutput="raw_values")
        tensor([0.5000, 0.7500])

        >>> input = torch.tensor([[0, 2], [1, 6]])
        >>> target = torch.tensor([[0, 1], [2, 5]])
        >>> r2_score(input, target, multioutput="variance_weighted")
        tensor(0.7000)

        >>> input = torch.tensor([1.2, 2.5, 3.6, 4.5, 6])
        >>> target = torch.tensor([1, 2, 3, 4, 5])
        >>> r2_score(input, target, multioutput="raw_values", num_regressors=2)
        tensor(0.6200)
    """

    _r2_score_param_check(multioutput, num_regressors)
    sum_squared_obs, sum_obs, sum_squared_residual, num_obs = _r2_score_update(
        input, target
    )
    return _r2_score_compute(
        sum_squared_obs,
        sum_obs,
        sum_squared_residual,
        num_obs,
        multioutput,
        num_regressors,
    )


def _r2_score_update(
    input: torch.Tensor,
    target: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    _r2_score_update_input_check(input, target)
    return _update(input, target)


@torch.jit.script
def _update(
    input: torch.Tensor,
    target: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    sum_squared_obs = torch.sum(torch.square(target), dim=0)
    sum_obs = torch.sum(target, dim=0)
    sum_squared_residual = torch.sum(torch.square(target - input), dim=0)
    num_obs = torch.tensor(target.size(0))
    return sum_squared_obs, sum_obs, sum_squared_residual, num_obs


def _r2_score_compute(
    sum_squared_obs: torch.Tensor,
    sum_obs: torch.Tensor,
    rss: torch.Tensor,
    num_obs: torch.Tensor,
    multioutput: str,
    num_regressors: int,
) -> torch.Tensor:
    if num_obs < 2:
        raise ValueError(
            "There is no enough data for computing. Needs at least two samples to calculate r2 score."
        )
    if num_regressors >= num_obs - 1:
        raise ValueError(
            "The `num_regressors` must be smaller than n_samples - 1, "
            f"got num_regressors={num_regressors}, n_samples={num_obs}.",
        )
    return _compute(
        sum_squared_obs,
        sum_obs,
        rss,
        num_obs,
        multioutput,
        num_regressors,
    )


@torch.jit.script
def _compute(
    sum_squared_obs: torch.Tensor,
    sum_obs: torch.Tensor,
    rss: torch.Tensor,
    num_obs: torch.Tensor,
    multioutput: str,
    num_regressors: int,
) -> torch.Tensor:
    tss = sum_squared_obs - torch.square(sum_obs) / num_obs
    # Calculate R2 score when multioutput is equal to raw_values.
    r_squared = 1 - (rss / tss)
    if multioutput == "uniform_average":
        r_squared = torch.mean(r_squared)
    elif multioutput == "variance_weighted":
        r_squared = torch.sum(r_squared * tss / torch.sum(tss))

    # If num_regressors is not equal to 0, adjusted R2 applies.
    if num_regressors != 0:
        r_squared = 1 - (1 - r_squared) * (num_obs - 1) / (num_obs - num_regressors - 1)
    return r_squared


def _r2_score_param_check(
    multioutput: str,
    num_regressors: int,
) -> None:
    if multioutput not in ("raw_values", "uniform_average", "variance_weighted"):
        raise ValueError(
            "The `multioutput` must be either `raw_values` or `uniform_average` or `variance_weighted`, "
            f"got multioutput={multioutput}."
        )
    if not isinstance(num_regressors, int) or num_regressors < 0:
        raise ValueError(
            "The `num_regressors` must an integer larger or equal to zero, "
            f"got num_regressors={num_regressors}."
        )


def _r2_score_update_input_check(
    input: torch.Tensor,
    target: torch.Tensor,
) -> None:
    if input.ndim >= 3 or target.ndim >= 3:
        raise ValueError(
            "The dimension `input` and `target` should be 1D or 2D, "
            f"got shapes {input.shape} and {target.shape}."
        )
    if input.size() != target.size():
        raise ValueError(
            "The `input` and `target` should have the same size, "
            f"got shapes {input.shape} and {target.shape}."
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

from typing import Iterable, Optional, TypeVar

import torch
from torcheval.metrics.functional.classification.precision import (
    _precision_compute,
    _precision_param_check,
    _precision_update,
)
from torcheval.metrics.metric import Metric

TPrecision = TypeVar("TPrecision")


class MulticlassPrecision(Metric[torch.Tensor]):
    """
    Compute the precision score, the ratio of the true positives and the sum of
    true positives and false positives.
    Its functional version is ``torcheval.metrics.functional.multi_class_precision``.
    We cast NaNs to 0 in case some classes have zero instances in the predictions.

    Args:
        num_classes:
            Number of classes.
        average:
            - ``'micro'``[default]:
                Calculate the metrics globally.
            - ``'macro'``:
                Calculate metrics for each class separately, and return their unweighted mean.
                Classes with 0 true and predicted instances are ignored.
            - ``'weighted'``"
                Calculate metrics for each class separately, and return their weighted sum.
                Weights are defined as the proportion of occurrences of each class in "target".
                Classes with 0 true and predicted instances are ignored.
            - ``None``:
                Calculate the metric for each class separately, and return
                the metric for every class.

    Example:
        >>> import torch
        >>> from torcheval.metrics import MulticlassPrecision
        >>> metric = MulticlassPrecision(num_classes=4)
        >>> input = torch.tensor([0, 2, 1, 3])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.5000)

        >>> metric = MulticlassPrecision(average=None, num_classes=4)
        >>> input = torch.tensor([0, 2, 1, 3])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor([1., 0., 0., 1.])

        >>> metric = MulticlassPrecision(average="macro", num_classes=2)
        >>> input = torch.tensor([0, 0, 1, 1, 1])
        >>> target = torch.tensor([0, 0, 0, 0, 1])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.5833)

        >>> metric = MulticlassPrecision(num_classes=4)
        >>> input = torch.tensor([[0.9, 0.1, 0, 0], [0.1, 0.2, 0.4, 0.3], [0, 1.0, 0, 0], [0, 0, 0.2, 0.8]])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.5)
    """

    def __init__(
        self: TPrecision,
        num_classes: Optional[int] = None,
        average: Optional[str] = "micro",
    ) -> None:
        super().__init__()
        _precision_param_check(num_classes, average)
        self.num_classes = num_classes
        self.average = average
        if average == "micro":
            self._add_state("num_tp", torch.tensor(0.0))
            self._add_state("num_fp", torch.tensor(0.0))
            self._add_state("num_label", torch.tensor(0.0))
        else:
            # num_classes has been verified as a positive integer. Add this line to bypass pyre.
            assert isinstance(
                num_classes, int
            ), f"num_classes must be a integer, but got {num_classes}"

            self._add_state("num_tp", torch.zeros(num_classes))
            self._add_state("num_fp", torch.zeros(num_classes))
            self._add_state("num_label", torch.zeros(num_classes))

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(
        self: TPrecision, input: torch.Tensor, target: torch.Tensor
    ) -> TPrecision:
        """
        Update states with the ground truth labels and predictions.

        Args:
            input: Tensor of label predictions.
                It could be the predicted labels, with shape of (n_sample, ).
                It could also be probabilities or logits with shape of (n_sample, n_class).
                ``torch.argmax`` will be used to convert input into predicted labels.
            target: Tensor of ground truth labels with shape of (n_sample, ).
        """
        num_tp, num_fp, num_label = _precision_update(
            input, target, self.num_classes, self.average
        )
        self.num_tp += num_tp
        self.num_fp += num_fp
        self.num_label += num_label
        return self

    @torch.inference_mode()
    def compute(self: TPrecision) -> torch.Tensor:
        """
        Return the precision score.

        0 is returned if no calls to ``update()`` are made before ``compute()`` is called.
        """
        return _precision_compute(
            self.num_tp, self.num_fp, self.num_label, self.average
        )

    @torch.inference_mode()
    def merge_state(self: TPrecision, metrics: Iterable[TPrecision]) -> TPrecision:
        for metric in metrics:
            self.num_tp += metric.num_tp.to(self.device)
            self.num_fp += metric.num_fp.to(self.device)
            self.num_label += metric.num_label.to(self.device)
        return self



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from typing import Tuple, Union

import torch


@torch.inference_mode()
def mean(
    input: torch.Tensor,
    weight: Union[float, int, torch.Tensor] = 1.0,
) -> torch.Tensor:
    """
    Compute weighted mean. When weight is not provided, it calculates the unweighted mean.
    Its class version is ``torcheval.metrics.Mean``.

    weighted_mean = sum(weight * input) / sum(weight)

    Args:
        input: Tensor of input values.
        weight(optional): Float or Int or Tensor of input weights. It is default to 1.0. If weight is a Tensor, its size should match the input tensor size.
    Raises:
        ValueError: If value of weight is neither a ``float`` nor a ``int`` nor a ``torch.Tensor`` that matches the input tensor size.

    Example:
        >>> import torch
        >>> from torcheval.metrics.functional import mean
        >>> mean(torch.tensor([2, 3]))
        tensor(2.5)
        >>> mean(torch.tensor([2, 3]), torch.tensor([0.2, 0.8]))
        tensor(2.8)
        >>> mean(torch.tensor([2, 3]), 0.5)
        tensor(2.5)
        >>> mean(torch.tensor([2, 3]), 1)
        tensor(2.5)
    """
    return _mean_compute(input, weight)


def _mean_update(
    input: torch.Tensor, weight: Union[float, int, torch.Tensor]
) -> Tuple[torch.Tensor, torch.Tensor]:
    if isinstance(weight, float) or isinstance(weight, int):
        weighted_sum = weight * torch.sum(input)
        weights = torch.tensor(float(weight) * torch.numel(input))
        return weighted_sum, weights
    elif isinstance(weight, torch.Tensor) and input.size() == weight.size():
        return torch.sum(weight * input), torch.sum(weight)
    else:
        raise ValueError(
            "Weight must be either a float value or a tensor that matches the input tensor size. "
            f"Got {weight} instead."
        )


def _mean_compute(
    input: torch.Tensor, weight: Union[float, int, torch.Tensor]
) -> torch.Tensor:
    weighted_sum, weights = _mean_update(input, weight)
    return weighted_sum / weights



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

from typing import Iterable, TypeVar, Union

import torch

from torcheval.metrics.functional.aggregation.sum import _sum_update
from torcheval.metrics.metric import Metric

TSum = TypeVar("TSum")


class Sum(Metric[torch.Tensor]):
    """
    Calculate the weighted sum value of all elements in all the input tensors.
    When weight is not provided, it calculates the unweighted sum.
    Its functional version is ``torcheval.metrics.functional.sum``.

    Example:
        >>> import torch
        >>> from torcheval.metrics import Sum
        >>> metric = Sum()
        >>> metric.update(1)
        >>> metric.update(torch.tensor([2, 3]))
        >>> metric.compute()
        tensor(6.)
        >>> metric.update(torch.tensor(-1)).compute()
        tensor(5.)
        >>> metric.reset()
        >>> metric.update(torch.tensor(-1)).compute()
        tensor(-1.)

        >>> metric = Sum()
        >>> metric.update(torch.tensor([2, 3]), torch.tensor([0.1, 0.6])).compute()
        tensor(2.)
        >>> metric.update(torch.tensor([2, 3]), 0.5).compute()
        tensor(4.5)
        >>> metric.update(torch.tensor([4, 6]), 1).compute()
        tensor(14.5)
    """

    def __init__(self: TSum) -> None:
        super().__init__()
        self._add_state("weighted_sum", torch.tensor(0.0))

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(
        self: TSum, input: torch.Tensor, weight: Union[float, int, torch.Tensor] = 1.0
    ) -> TSum:
        """
        Update states with the values and weights.

        Args:
            input: Tensor of input values.
            weight(optional): Float or Int or Tensor of input weights. It is default to 1.0. If weight is a Tensor, its size should match the input tensor size.
        Raises:
            ValueError: If value of weight is neither a ``float`` nor ``int`` nor a ``torch.Tensor`` that matches the input tensor size.
        """

        self.weighted_sum += _sum_update(input, weight)
        return self

    @torch.inference_mode()
    def compute(self: TSum) -> torch.Tensor:
        return self.weighted_sum

    @torch.inference_mode()
    def merge_state(self: TSum, metrics: Iterable[TSum]) -> TSum:
        for metric in metrics:
            self.weighted_sum += metric.weighted_sum.to(self.device)
        return self



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from torcheval.metrics.functional.ranking.hit_rate import hit_rate
from torcheval.metrics.functional.ranking.reciprocal_rank import reciprocal_rank

__all__ = [
    "hit_rate",
    "reciprocal_rank",
]
